# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
# from hbmqtt.errors import BrokerException
#
# class HandlerException(BrokerException):
#     """
#     Exceptions thrown by encode/decode process
#     """
#     pass
#
